#!/usr/bin/perl

# exceptional driver program for Web-Learning-Project
# (c) 1995 by Burt Rosenberg. All rights reserved.
#

# called with ?~ip0 and will open ip0.
# Generally looks at env. var. CLICK
# to determine which page to open.

# configuration done by wlp.conf files found in user directory, along
#   with data files (contexts and digests)

# digests contain various forms.
# forms contain html and directives that authorize the formation of
#   radio buttons.
# each radio buttons is connected to a message, called CLICK, which
#   contains the form name and an argument, dot separated.

  $Debug = 0 ;  # non-zero for debugging ON
  $ErrorMsg = 1 ; # non-zero for report of error messages
  $pathname = "~burt/projets/wlp" ;
  $contextname = "burt" ;
  $wirefilename = "wirefile" ;
  $contentsfilename = "contents" ;
  $workdir = "tmp" ;
  $error_msg = "" ;  # Empty error message means no errors.
  $Session_Key = "" ;

  $0 =~ /([\w\.]*)$/ ;
  $drivername = $1 ;
  srand(time|$$) ;

  # if called with an explicit page name as argument
  if ( $ENV{'QUERY_STRING'} =~ /^~(.*)/ ) {
     $currentForm = $1 ;
     $Session_Key = $contextname . "." . &CreateSessionKey ;
  }

  &html_header ;
  &ReadParse ;
  # Map click using  context_file and wire_file 

  if ($in{'CLICK'}) {
     # the question: Is there a CLICK? is equiv. to: Is there a SESSION_KEY?
     # the SESSION_KEY contains the directory (context) in which other
     # data (wirefile name, etc) is found, directory/wlp.conf. (future plans!)
 
     # CLICKS are mutlipart: message.message,flags
     # burst out click_flags
     if ($in{'CLICK'} =~ /,/) {
        $in{'CLICK'} =~ /(.*),(.*)/ ;
        $click_flags = $2 ;
        $in{'CLICK'} = $1 ;
     }
     else {
        $click_flags = "" ;
     }

     $Session_Key = $in{'SESSION_KEY'} ;
     unless ( $currentForm = &MapInContextFile( $in{'CLICK'}, $Session_Key )) {
        $currentForm = &MapInWireFile( $in{'CLICK'} ) ;
     }
  }

  # get and open document which contains the form
  if ( !$currentForm ) {
     print "<pre>I could not locate Form |$currentForm|.</pre>" ;
     exit ;
  }
  $currentDig = &MapInContentsFile( $currentForm ) ;
  unless ( $currentDig ) {
     print "<pre>I could not locate Digest associated with Form |$currentForm|.</pre>" ;
     exit ;
  }
  if ( !open( DIGEST, "$pathname/$currentDig" ) ) {
     print "<pre>I could not open Digest |$currentDig| associated " ;
     print "with Form |$currentForm|.</pre>" ;
     exit ;
  }

  # predefine macros
  $macros{'SESSION_KEY'} = $Session_Key ;
  $macros{'FORM'} = $currentForm ;
  $macros{'DATE'} = `date` ;

  &ProcessInput( $currentForm ) ;

  if ( $click_flags =~ /D/ ) { # this is a digression
    $error_msg .=  &PushMessage( $Session_Key, $in{'CLICK'} ) ;
  }

  $error_msg .= "click_flags:|$click_flags|\n" ;
  if ( $ErrorMsg ) {
    printf "<pre>Error messages:\n" ;
    print  "$error_msg</pre>" ;
  }
  close(DIGEST) ;

# ---------------------------- end main line

sub html_header {
    print "Content-type: text/html\n\n";
}

sub CreateSessionKey {
  local ($id) = `date` ;
  $id =~ s/ //g ;
  $id =~ s/://g ;
  substr($id, 3, 11 ) ;
}

sub PushMessage {
    local( $sk, $ck ) = ( $_[0], $_[1] ) ;
    # sk is the sessions key, generally of the form context.uniqueKey
    # ck is the key click, usually of the form fromPage.message
    # append this information to a context file

    $sk =~ /.*\.(.*)/ ;
    local( $uk ) = $1 ;
    $ck =~ /(.*)\..*/ ;
    local( $pg ) = $1 ;

    unless ( open(OUTFILE, ">$workdir/$uk.tmp") ) {
       return "$workdir/$uk" ;
    }
    printf OUTFILE "*.%s:%s\n", $uk, $pg ;
    if ( open(INFILE, "$workdir/$uk") ) {
       while (<INFILE>) { print OUTFILE ; }
       close(INFILE) ;
    }

    close(OUTFILE) ;
    unlink ("$workdir/$uk") ;
    rename ("$workdir/$uk.tmp","$workdir/$uk") ;
    return "" ;
}

sub MapInWireFile {

   # finds the form associated with the CLICK message.
   # messages are dot-separated text, usually page_name.message_option,
   # the wirefile contains a matching p_n.m_o:form_to_get, or p_n, m_o
   # can be "*", for a wildcard match.

   # returns the form name if successful. Else returns the empty string.

   if ( !open ( WIREFILE, "$pathname/$wirefilename" ) ) {
      $error_msg .= "MapInWireFile $wirefilename open failure.\n" ;
      return "" ;
   }
   # wire file suc. opened.

   local( $click ) = $_[0] ;
   local( $np ) = "" ;
   local ( @oparts, $i, $j ) ;
   local ( @ckparts ) = split( /\./, $click ) ; 

   while (<WIREFILE>) {
     /^#/ && next ;
     /(.*):(.*)/ ;
     $j = $2 ;
     @oparts = split( /\./, $1 ) ;
     if ( $#oparts != $#ckparts ) { next ; }
     for ( $i=0; $i<=$#ckparts; $i++) {
        if ( ($oparts[$i] ne $ckparts[$i]) &&
             ($oparts[$i] ne "*") ) { last ; }
     }
     if ( $i > $#ckparts ) { 
        $np = $j ;
        last ;
     }
   } # elihw
   close( WIREFILE ) ;

   if ($np eq "" ) {
     $error_msg .= "MapInWireFile $click lookup failure.\n" ;
   }
   $np ;
}

sub MapInContextFile {    # click, session_key

   # finds the form associated with the CLICK message.
   # messages are dot-separated text, usually page_name.message_option,
   # the wirefile contains a matching p_n.m_o:form_to_get, or p_n, m_o
   # can be "*", for a wildcard match.

   # returns the form name if successful. Else returns the empty string.

   local( $click ) = $_[0] ;
   unless ( $_[1] ) { "" ; }  # Return if no session key.
   local ( $sk ) = $_[1] ;   # arg1, the $Session_Key ;
   $sk =~ /.*\.(.*)/ ;
   local ( $uk ) = $1 ;

   unless ( open ( CONTEXTFILE, "$workdir/$uk" ) ) {
      $error_msg .= "MapInContextFile $workdir/$uk open failure.\n" ;
      return "" ;
   }
   unless ( open ( CONTEXTTEMP, ">$workdir/$uk.tmp" ) ) {
      $error_msg .= "MapInContextFile $workdir/$uk.tmp open failure.\n" ;
      return "" ;
   }

   local( $np ) = "" ;
   local ( @oparts, $i, $j ) ;
   local ( @ckparts ) = split( /\./, $click ) ; 

   while (<CONTEXTFILE>) {
     if (/^#/) {
        print CONTTEXTTEMP ;
        next ;
     }
     /(.*):(.*)/ ;
     $j = $2 ;
     @oparts = split( /\./, $1 ) ;
     if ( $#oparts != $#ckparts ) { 
        print CONTEXTTEMP ;
        next ;
     }
     for ( $i=0; $i<=$#ckparts; $i++) {
        if ( ($oparts[$i] ne $ckparts[$i]) &&
             ($oparts[$i] ne "*") ) { last ; }
     }
     if ( $i > $#ckparts ) { 
        $np = $j ;
        last ;
     }
     print CONTEXTTEMP ;
   } # elihw


   while (<CONTEXTFILE>) { print CONTEXTTEMP ; }
   close( CONTEXTFILE ) ;
   close( CONTEXTTEMP ) ;

   if ( $np eq "" ) {
     unlink ( "$workdir/$uk.tmp" ) ;
   }
   else {
     unlink( "$workdir/$uk" ) ;
     rename( "$workdir/$uk.tmp", "$workdir/$uk" ) ;
   }

   $np ;
}

sub MapInContentsFile {

   # find the digest of forms which contains the form $_[0]
   # the contentsfile is (currently) maintained externally by a user run routine.

   unless ( open ( CONTENTSFILE, "$pathname/$contentsfilename" ) ) {
      print "MapInContentsFile $pathname/$contentsfilename open failure." ;
      $error_msg .= "MapInContentsFile $contentsfilename open failure." ;
      return "" ;
   }

   local( $target ) = $_[0] ;
   local ( $np ) = "" ;

   while (<CONTENTSFILE>) {
     /^#/ && next ;
     if (/(.*):\s*$target/) {
       $np = $1 ;
       last ;
      }
   }
   close (CONTENTSFILE) ;
   if ($np eq "" ) {
     $error_msg .= "MapInContentsFile $target lookup failure." ;
   }
   $np ;
}

# ------------ Processing of a form ---------------------

sub getDigest {
  # if the argument is "", get next line, else
  # look for #Form: $_[0] line in DIGEST.
  # return empty line if done.

  local ( $line ) ;

  if ( $_[0] ) {
     do {
       $line = <DIGEST> ;
       $line =~ /^#Form:\s*(\w*)\s*(.*)/ ;
     } until $1 eq $_[0] ;
  }
  else {
     do {
        $line = <DIGEST> ;
        if ( $line =~ /^#Form:/ ) {
           $line = "" ;
        }
     }
     until ! ($line =~ /^# /) ;
  }
  $line ;
}

sub ProcessInput {
  local ( $theForm ) = $_[0] ;
  local ( $title ) ;

  $line = &getDigest( $theForm ) ;

  if ( $line ) {
     # process form

     $line =~ /^#Form:\s*(\w*)\s*(.*)/ ;
     $thePage = $1 ;
     printf "<title>$2</title>\n" ;
     printf '<FORM METHOD="POST" ACTION="%s">', $drivername ;
     printf "\n" ;
     printf '<input type="hidden" name="SESSION_KEY" value="%s">', $Session_Key ;
     printf "\n" ;
     &prodForm ;
     printf "<p><input type=\"submit\" value=\"Enter\" ></FORM>" ;
     printf "<em>(c) Copyright 1995 Burt Rosenberg. All rights reserved.</em>" ;
   }

   else {
     # else verbatim it.
     print $line ;
     while (<DIGEST>) {
        print $_ ;
     }
   }
}

sub prodForm {
  # Produce a form
  while ($_ = &getDigest("")) {
     if ( /^#Shuffle:(.*)/ ) {
        &DoShuffle($1) ;
     }
     elsif ( /^#List:(.*)/ ) {
        &DoList( $1 ) ;
     }
     elsif ( /^#Digress:(.*)/ ) {
        &DoDigress( $1 ) ;
     }
     elsif ( /^#Return:(.*)/ ) {
        &DoReturn( $1 ) ;
     }
     elsif ( /^#Macro: (\w*)/ ) {
        &DoMacro( $1, $' ) ;
     }
     elsif ( /^#Eval: (.*)/ ) {
        &DoEval($1) ;
     }
     else { print $_ ; }
  }
}
  
sub DoEval {
  print &MacroExp($_[0]) ;
}

sub MacroExp {
  local ( $line ) = @_ ;
  local ( $s ) ;
  while ( $line =~ /\$(\w+)\$/ ) {
     $s = $1 ;
     if ( $macros{$s} ) {
        $line =~ s/\$$s\$/$macros{$s}/ ; 
     }
     else {
         $line =~ s/\$$s\$/$s/ ;
     }
  }
  return $line ;
}

sub DoMacro {
  local ( $name, $value ) = @_ ;
  $value =~ /\s*(.*\S)/ ; #strip leading an trailing whitespace
  $macros{$name} = $1;
  if ( $Debug ) {
     printf "<Pre>Debug: Macro arguments\n" ;
     printf "  name=|%s|, value=|%s|\n</pre>", $name, $macros{$name} ;
  }
}

sub DoPuzzle {
  # directive #Puzzle: block0
  # each line of puzzle ending with #:

  local ( @args ) = split(' ',&MacroExp($_[0])) ;
  local ($i, $j ) ;
  local ( $pi, $origpi ) ;
  local ( %t ) ;
  # how do I declare t local?

  if ( $Debug ) {
     printf "<PRE>Debug: Puzzle arguments:\n" ;
     for ( $i=0;$i<=$#args;$i++ ) {
        printf "   %d, |%s|\n", $i, $args[$i] ;
     }
     printf "</PRE>" ;
  }

  # read the blocks of text.
  $i = 0 ;
  $j = 'A' ;
  $pi = $j ;
  while (<DIGEST>) {
     /#:$args[0]/ && last ;
     if (/^#:/ ) {
         $i++ ; 
         $j = sprintf("%c", ord('A')+$i) ;
         $pi .= $j ;
         $t{$j} = "" ;
     }
     else { $t{$j} .= $_ ; }
  }
  $origpi = $pi ;
  # postcondition $t{A..T} has text blocks, $pi=A..T

  if ( $in{'THE_PUZZLE'} ) {
     # not the first time thru, handle things ...
     $in{'THE_PUZZLE'} =~ /^(.*)-(.*)/ ;
     $pi = $2 ;
     if ( $Debug ) {
        printf "<pre>Debug: Puzzle rcvd data,\n" ;
        printf "   THE_PUZZLE=|%s|, blockname=|%s|, permutation=|%s|\n",
           $in{'THE_PUZZLE'}, $1, $2 ;
        printf "   THE_DATA=|%s| length %d </pre>", $in{'THE_DATA'},
           length $in{'THE_DATA'} ;
     }
     # now transpose according to THE_DATA
     if ( $in{'THE_DATA'} =~ /(\w)$;(\w)/ ) {

       local($a) = $1 ;
       local($b) = $2 ;
       local($c, $newpi) ;

       while ( $pi ) {
          $c = substr( $pi, 0, 1 ) ;
          if ( $c eq $a ) { $newpi .= $b ; }
          elsif ( $c eq $b )  { $newpi .= $a ; }
          else {
            $newpi .= $c ; }
          substr( $pi, 0, 1 ) = "" ;
       }
       $pi = $newpi ;
     }
  }
  else {
     # first time thru, generate a shuffle of $pi
     local($newpi) ;
     while ( $pi ) {
        $j = rand(length $pi ) ;
        $newpi .= substr( $pi, $j, 1 ) ;
        substr( $pi, $j, 1 ) = '' ;
     }
     $pi = $newpi ;
  }
  #post: $pi is the current permutation

  if ( $origpi eq $pi ) {
     # puzzle is solved
     printf '<h2>Solved!</h2><pre>' ;
     while ( $pi ) {
        $j = substr( $pi, 0, 1 ) ;
        substr( $pi, 0, 1 ) = "" ;
        printf "%s", $t{$j} ;
     }
     printf "</pre>" ;
  }
  else {
     printf '<FORM METHOD="POST" ACTION="$drivername.pl">' ;
     printf '<input type="hidden" name="THE_PAGE" value="%s">', $page ;
     printf '<input type="hidden" name="THE_PUZZLE" value="%s-%s">', 
        $args[0], $pi ;

     while ( $pi ) {
        $j = substr( $pi, 0, 1 ) ;
        substr( $pi, 0, 1 ) = "" ;
        printf '<br><input type="checkbox" name="THE_DATA" value="%s">', $j ;
        printf '%s', $t{$j} ;
     }
     printf '<p><input type="submit" value="Enter" ></FORM>' ;
  }
}

sub DoShuffle {
  # directive #Shuffle: block0 block1 ...
  # each block ended with #: 
  # only in a #Form:

  local ( @args ) = split(' ',&MacroExp($_[0])) ;
  local ($i, $j, @block, @tmpblock ) ;

  if ( $Debug) {
     printf "<PRE>Debug: Shuffle arguments:\n" ;
     for ( $i=0;$i<=$#args;$i++ ) {
        printf "   %d, |%s|\n", $i, $args[$i] ;
     }
     printf "</PRE>" ;
  }

  # read the blocks of text.
  for ( $i=0;$i<=$#args;$i++ ) {
     $block[$i] = $args[$i] ;
     while ($_ = &getDigest("")) {
        /^#:/ && last ;
        $block[$i] .= "\n" . $_ ;
     } 
  }

  # shuffle them,
  # page 243 Llama book
  push ( @tmpblock, splice( @block, rand(@block),1)) while @block ;
  @block = @tmpblock ;

  print "<menu>" ;
  # generate buttons
  # in the next for, it's supposed to be strict <
  for ( $i=0;$i<=$#args;$i++ ) {
     $block[$i] =~ /(.*)/ ;
     printf '<br><input type="radio" name="CLICK" value="%s.%s">', 
       $thePage, $1 ;
     printf "%s", $' ;
  }
  print "</menu>" ;
}

sub DoList {
  # directive #List: block0 block1 ...
  # each block ended with #:
  # supported only in a #Form:

  local ( @args ) = split(' ',&MacroExp($_[0])) ;
  local ($i, $checked ) ;

  if ( $Debug) {
     printf "<PRE>Debug: List arguments:\n" ;
     for ( $i=0;$i<=$#args;$i++ ) {
        printf "   %d, |%s|\n", $i, $args[$i] ;
     }
     printf "</PRE>" ;
  }

  print "<menu>" ;
  # read the blocks of text.
  for ( $i=0;$i<=$#args;$i++ ) {
     
     # decide if this is a checked button
     $checked = "" ;
     if ( $args[$i] =~ /\*(.*)/ ) {
        $args[$i] = $1 ; # remove the star
        $checked = "CHECKED" ;
     }
     printf '<br><input type="radio" name="CLICK" value="%s.%s" %s>',
       $thePage, $args[$i], $checked ;
     
     while ($_ = &getDigest("")) {
        /^#:/ && last ;
        printf "%s", $_ ;
    }
  }
  print "</menu>" ;
}

sub DoReturn {
  # directive #Return:
  # block ends with #:

  local ( @args ) = split(' ',&MacroExp($_[0])) ;
  local ( $checked ) ;

  print "<menu>" ;
  $Session_Key =~ /.*\.(.*)/ ;
  local($uk) = $1 ;

  # decide if this is a checked button
  $checked = "" ;
  if ( $args[$0] =~ /\*/ ) {
     $checked = "CHECKED" ;
  }

  printf '<br><input type="radio" name="CLICK" value="%s.%s" %s>',
     $thePage, $uk, $checked ;
     
  while ($_ = &getDigest("")) {
     /^#:/ && last ;
     printf "%s", $_ ;
  }
  print "</menu>" ;
}

sub DoDigress {
  # directive #Digress: block0 block1 ...
  # each block ended with #:
  # supported only in a #Form:

  local ( @args ) = split(' ',&MacroExp($_[0])) ;
  local ($i, $checked ) ;

  if ( $Debug) {
     printf "<PRE>Debug: List arguments:\n" ;
     for ( $i=0;$i<=$#args;$i++ ) {
        printf "   %d, |%s|\n", $i, $args[$i] ;
     }
     printf "</PRE>" ;
  }

  print "<menu>" ;
  # read the blocks of text.
  for ( $i=0;$i<=$#args;$i++ ) {
     
     # decide if this is a checked button
     $checked = "" ;
     if ( $args[$i] =~ /\*(.*)/ ) {
        $args[$i] = $1 ; # remove the star
        $checked = "CHECKED" ;
     }
     printf '<br><input type="radio" name="CLICK" value="%s.%s,D" %s>',
       $thePage, $args[$i], $checked ;
     
     while ($_ = &getDigest("")) {
        /^#:/ && last ;
        printf "%s", $_ ;
    }
  }
  print "</menu>" ;
}

# ------------------ additional code


# Source: cgi-lib.pl, httpd distribution.
# Copyright 1993 Steven E. Brenner

sub ReadParse {
  if (@_) { local (*in) = @_; }

  local ($i, $loc, $key, $val);

  # Read in text
  if ($ENV{'REQUEST_METHOD'} eq "GET") {
    $in = $ENV{'QUERY_STRING'};
  } elsif ($ENV{'REQUEST_METHOD'} eq "POST") {
    for ($i = 0; $i < $ENV{'CONTENT_LENGTH'}; $i++) {
      $in .= getc;
    }
  }

  @in = split(/&/,$in);

  foreach $i (0 .. $#in) {
    # Convert plus's to spaces
    $in[$i] =~ s/\+/ /g;
   
    # Convert %XX from hex numbers to alphanumeric
    $in[$i] =~ s/%(..)/pack("c",hex($1))/ge;

    # Split into key and value.
    $loc = index($in[$i],"=");
    $key = substr($in[$i],0,$loc);
    $val = substr($in[$i],$loc+1);
    $in{$key} .=  $; if (defined($in{$key})); # \0 is the multiple separator
    $in{$key} .= $val;
  }

  return 1; 
}

sub PrintVariables {
  local (%in) = @_;
  local ($old, $out);
  $old = $*;  $* =1;
  $output .=  "<DL COMPACT>";
  foreach $key (sort keys(%in)) {
    ($out = $in{$key}) =~ s/\n/<BR>/g;
    $output .=  "<DT><B>$key</B><DD><i>$out</I><BR>";
  }
  $output .=  "</DL>";
  $* = $old;

  return $output;
}
  

