#!/usr/local/bin/perl

# new driver program for Web-Learning-Project
# implements the meta-wlp castings
#
# (c) burt rosenberg 1995, free to distribute
# 11 Feb 1995
#
# if file DEBUG exists in this directory, then
# just report on the variables returned.
#

  $pathname = "vm" ;
  $wirefile = "wire" ;
  $drivername = "vmdri" ;
  $homepage = "$pathname/hello.html" ;

  srand(time) ;
  $Debug = 0 ;  # non-zero for debugging ON

  &html_header('Title of page') ;
  &ReadParse ;

  if ( $in{'Chapter'} ) {
    #  start of a chapter
    $page = $in{'Chapter'} ;
    if ( $in{'Exercises'} eq "yes" ) {
       $page .= "-X" ;
    }
    else {
       $page .= "-T" ;
    }
    $page = &MapInWireFile( $page ) ;
    $file = "$pathname/$page.html" ;
    open( DOCUMENT, $file ) ||
       die "<PRE>$0: $file not found\n</PRE>" ;
    &ProcessInput ;
    close(DOCUMENT) ;
  }

  elsif ( $in{'THE_PUZZLE'} ) {
    # Returning from a puzzle page.
    $page = $in{'THE_PAGE'} ;
    #reopen page
    $file = "$pathname/$page.html" ;
    open( DOCUMENT, $file ) ||
       die "<PRE>$0: $file not found\n</PRE>" ;
    &ProcessInput ;
    close(DOCUMENT) ;
  }

  elsif ( $in{'Guess'} ) {
     #  handle the guess
     $guess = $in{'Guess'} ;
     $np = &MapInWireFile( $guess ) ;
     if ( $np ne "" ) {
       $file = "$pathname/$np.html" ;
       open( DOCUMENT, $file ) ||
          die "<PRE>$0: $file not found\n</PRE>" ;
       &ProcessInput ;
       close(DOCUMENT) ;
     }
  }

  else {
    open( DOCUMENT, $homepage ) ||
         die "<PRE>$0: $homepage not found\n</PRE>" ;
    &ProcessInput ;
    close(DOCUMENT) ;
#    &PrintVariables(%in) ;
#    print $output;
  }

  &html_trailer ;

# end main line

sub MapInWireFile {
   local( $guess ) = $_[0] ;
   local( $np ) ;

   open ( WIREFILE, "$pathname/$wirefile" ) ||
       die "<PRE>$0: $pathname/$wirefile not found\n</PRE>" ;
   while (<WIREFILE>) {
     /^#/ && next ;
     if (/$guess:(.*)/) {
       $np = $1 ;
       last ;
      }
   } # elihw
   if ($np eq "" ) {
     printf "<PRE>$0: translation of $guess failed.\n</PRE>" ;
   }
   $np ;
}

sub ProcessInput {
  $line = <DOCUMENT> ;
  if ( $line =~ /^#Meta-Wlp:/ ) {
    &metawlp ; }
  else {
     # else if not MetaWlp, verbatim it.
     print $line ;
     while (<DOCUMENT>) {
        print $_ ;
     }
   }
}

sub metawlp {
  # Process Meta-Wlp, header line already read.
  local($i) ;

  while (<DOCUMENT>) {
     if ( /^#Shuffle:(.*)/ ) {
        &DoShuffle($1) ;
     }
     elsif ( /^#Macro: (\w*)/ ) {
        &DoMacro( $1, $' ) ;
     }
     elsif ( /^#Eval: (.*)/ ) {
        &DoEval($1) ;
     }
     elsif ( /^#Puzzle: (.*)/) {
        &DoPuzzle($1) ;
     }
     else { print $_ ; }
  }
}

sub DoEval {
  print &MacroExp($_[0]) ;
}

sub MacroExp {
  local ( $line ) = @_ ;
  local ( $s ) ;
  while ( $line =~ /\$(\w+)\$/ ) {
     $s = $1 ;
     if ( $macros{$s} ) {
        $line =~ s/\$$s\$/$macros{$s}/ ; 
     }
     else {
         $line =~ s/\$$s\$/$s/ ;
     }
  }
  return $line ;
}

sub DoMacro {
  local ( $name, $value ) = @_ ;
  $value =~ /\s*(.*\S)/ ; #strip leading an trailing whitespace
  $macros{$name} = $1;
  if ( $Debug ) {
     printf "<Pre>Debug: Macro arguments\n" ;
     printf "  name=|%s|, value=|%s|\n</pre>", $name, $macros{$name} ;
  }
}

sub DoPuzzle {
  # directive #Puzzle: block0
  # each line of puzzle ending with #:

  local ( @args ) = split(' ',&MacroExp($_[0])) ;
  local ($i, $j ) ;
  local ( $pi, $origpi ) ;
  local ( %t ) ;
  # how do I declare t local?

  if ( $Debug ) {
     printf "<PRE>Debug: Puzzle arguments:\n" ;
     for ( $i=0;$i<=$#args;$i++ ) {
        printf "   %d, |%s|\n", $i, $args[$i] ;
     }
     printf "</PRE>" ;
  }

  # read the blocks of text.
  $i = 0 ;
  $j = 'A' ;
  $pi = $j ;
  while (<DOCUMENT>) {
     /#:$args[0]/ && last ;
     if (/^#:/ ) {
         $i++ ; 
         $j = sprintf("%c", ord('A')+$i) ;
         $pi .= $j ;
         $t{$j} = "" ;
     }
     else { $t{$j} .= $_ ; }
  }
  $origpi = $pi ;
  # postcondition $t{A..T} has text blocks, $pi=A..T

  if ( $in{'THE_PUZZLE'} ) {
     # not the first time thru, handle things ...
     $in{'THE_PUZZLE'} =~ /^(.*)-(.*)/ ;
     $pi = $2 ;
     if ( $Debug ) {
        printf "<pre>Debug: Puzzle rcvd data,\n" ;
        printf "   THE_PUZZLE=|%s|, blockname=|%s|, permutation=|%s|\n",
           $in{'THE_PUZZLE'}, $1, $2 ;
        printf "   THE_DATA=|%s| length %d </pre>", $in{'THE_DATA'},
           length $in{'THE_DATA'} ;
     }
     # now transpose according to THE_DATA
     if ( $in{'THE_DATA'} =~ /(\w)$;(\w)/ ) {

       local($a) = $1 ;
       local($b) = $2 ;
       local($c, $newpi) ;

       while ( $pi ) {
          $c = substr( $pi, 0, 1 ) ;
          if ( $c eq $a ) { $newpi .= $b ; }
          elsif ( $c eq $b )  { $newpi .= $a ; }
          else {
            $newpi .= $c ; }
          substr( $pi, 0, 1 ) = "" ;
       }
       $pi = $newpi ;
     }
  }
  else {
     # first time thru, generate a shuffle of $pi
     local($newpi) ;
     while ( $pi ) {
        $j = rand(length $pi ) ;
        $newpi .= substr( $pi, $j, 1 ) ;
        substr( $pi, $j, 1 ) = '' ;
     }
     $pi = $newpi ;
  }
  #post: $pi is the current permutation

  if ( $origpi eq $pi ) {
     # puzzle is solved
     printf '<h2>Solved!</h2><pre>' ;
     while ( $pi ) {
        $j = substr( $pi, 0, 1 ) ;
        substr( $pi, 0, 1 ) = "" ;
        printf "%s", $t{$j} ;
     }
     printf "</pre>" ;
  }
  else {
     printf '<FORM METHOD="POST" ACTION="/wlp/$drivername.pl">' ;
     printf '<input type="hidden" name="THE_PAGE" value="%s">', $page ;
     printf '<input type="hidden" name="THE_PUZZLE" value="%s-%s">', 
        $args[0], $pi ;

     while ( $pi ) {
        $j = substr( $pi, 0, 1 ) ;
        substr( $pi, 0, 1 ) = "" ;
        printf '<br><input type="checkbox" name="THE_DATA" value="%s">', $j ;
        printf '%s', $t{$j} ;
     }
     printf '<p><input type="submit" value="Enter" ></FORM>' ;
  }
}

sub DoShuffle {
  # directive #Shuffle: block0 block1 ...
  # each block ended with #: 

  local ( @args ) = split(' ',&MacroExp($_[0])) ;
  local ($i, $j, @block, @tmpblock ) ;

  if ( $Debug) {
     printf "<PRE>Debug: Shuffle arguments:\n" ;
     for ( $i=0;$i<=$#args;$i++ ) {
        printf "   %d, |%s|\n", $i, $args[$i] ;
     }
     printf "</PRE>" ;
  }

  # read the blocks of text.
  for ( $i=0;$i<=$#args;$i++ ) {
     $block[$i] = $args[$i] ;
     while (<DOCUMENT>) {
        /^#:/ && last ;
        $block[$i] .= "\n" . $_ ;
     } 
  }

  # shuffle them,
  # page 243 Llama book
  push ( @tmpblock, splice( @block, rand(@block),1)) while @block ;
  @block = @tmpblock ;

  # generate form
  printf "<FORM METHOD=\"POST\" ACTION=\"/wlp/$drivername.pl\">" ;
  # in the next for, it's supposed to be strict <
  for ( $i=0;$i<=$#args;$i++ ) {
     $block[$i] =~ /(.*)/ ;
     if ( $1 eq $args[0] ) { $j = $i ; next ; }
     printf "<br><input type=\"radio\" name=\"Guess\" value=\"%s\">", 
       $1 ;
     printf "%s", $' ;
  }

  $block[$j] =~ /(.*)/ ;
  printf "<p><input type=\"radio\" name=\"Guess\" value=\"" ;
  printf "%s\" CHECKED> %s", $1, $';
  printf "<p><input type=\"submit\" value=\"Enter\" ></FORM>" ;

}

# Source: cgi-lib.pl, httpd distribution.
# Copyright 1993 Steven E. Brenner

sub ReadParse {
  if (@_) { local (*in) = @_; }

  local ($i, $loc, $key, $val);

  # Read in text
  if ($ENV{'REQUEST_METHOD'} eq "GET") {
    $in = $ENV{'QUERY_STRING'};
  } elsif ($ENV{'REQUEST_METHOD'} eq "POST") {
    for ($i = 0; $i < $ENV{'CONTENT_LENGTH'}; $i++) {
      $in .= getc;
    }
  }

  @in = split(/&/,$in);

  foreach $i (0 .. $#in) {
    # Convert plus's to spaces
    $in[$i] =~ s/\+/ /g;
   
    # Convert %XX from hex numbers to alphanumeric
    $in[$i] =~ s/%(..)/pack("c",hex($1))/ge;

    # Split into key and value.
    $loc = index($in[$i],"=");
    $key = substr($in[$i],0,$loc);
    $val = substr($in[$i],$loc+1);
    $in{$key} .=  $; if (defined($in{$key})); # \0 is the multiple separator
    $in{$key} .= $val;
  }

  return 1; # just for fun
}

sub PrintVariables {
  local (%in) = @_;
  local ($old, $out);
  $old = $*;  $* =1;
  $output .=  "<DL COMPACT>";
  foreach $key (sort keys(%in)) {
    ($out = $in{$key}) =~ s/\n/<BR>/g;
    $output .=  "<DT><B>$key</B><DD><i>$out</I><BR>";
  }
  $output .=  "</DL>";
  $* = $old;

  return $output;
}
  
# Source: cgi_handlers.pl, httpd distribution.
# James Tappin: sjt@xun8.sr.bham.ac.uk
# School of Physics & Space Research University of Birmingham
# Feb 1993.
  
sub html_header {
    local($title) = @_;
    print "Content-type: text/html\n\n";
#    print "<html><head>\n";
#    print "<title>$title</title>\n";
#    print "</head>\n<body>\n";
}

sub html_trailer {
    local($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst)
        = gmtime;
    local($mname) = ("Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
                     "Aug", "Sep", "Oct", "Nov", "Dec")[$mon];
    local($dname) = ("Sun", "Mon", "Tue", "Wed", "Thu", "Fri",
                     "Sat")[$wday];
#     print "<p>\nGenerated by: <var>$0</var><br>\n";
#    print "Date: $hour:$min:$sec UT on $dname $mday $mname $year.<p>\n";
    print "</body></html>\n";
}

